import json

from twisted.web import static
from twisted.web.resource import Resource

from ostrich import stats
from ostrich.time_series_collector import TimeSeriesCollector
from ostrich.timing import TimingStat

def respond(request, data, code=200):
    data = json.dumps(data, default=stats.json_encoder)
    if 'callback' in request.args:
        request.setResponseCode(200)
        request.setHeader('Content-Type', 'application/javascript')
        return request.args['callback'][0] + "(" + data + ")\n"
    else:
        request.setResponseCode(code)
        request.setHeader('Content-Type', 'application/json')
        return data + "\n"

class StatsResource(Resource):
    def render_GET(self, request):
        reset = int(request.args.get('reset', [0])[0])
        data = stats.stats(reset=reset)
        return respond(request, data)

class StatsTimeSeriesResource(StatsResource):
    def __init__(self, collect_every=60):
        Resource.__init__(self)
        self.collector = TimeSeriesCollector()
        self.collector.start_twisted(collect_every=collect_every)
        
        self.putChild('graph', static.Data(GRAPH_HTML.strip(), "text/html"))
        self.putChild('graph_data', TimeSeriesDataResource(self.collector))
        self.putChild('combined', TimeSeriesCombinedResource(self.collector))
    
    # def render_GET(self, request):
    #     reset = int(request.args.get('reset', [0])[0])
    #     return json.dumps(stats.stats(reset=reset), default=stats.json_encoder)

class TimeSeriesDataResource(Resource):
    isLeaf = True
    
    def __init__(self, collector):
        self.collector = collector
    
    def render_GET(self, request):
        if len(request.postpath) == 0:
            return respond(request, {'keys': self.collector.keys(), 'stats': self.collector.stats.stats()})
        else:
            def convert(v):
                if isinstance(v, TimingStat):
                    return v.to_dict()
                return v

            name = '/'.join(request.postpath)

            output = {}
            try:
                for n in name.split(','):
                    output[n] = [(date, convert(value)) for date, value in self.collector.get(n)]
            except KeyError:
                return respond(request, dict(code=404, error="Not Found"), code=404);

            if ',' in name:
                return respond(request, output)
            else:
                return respond(request, output[name])

class TimeSeriesCombinedResource(Resource):
    isLeaf = True
    
    def __init__(self, collector):
        self.collector = collector
    
    def render_GET(self, request):
        series = request.args.get('series', ['false'])[0] == 'true'
        if len(request.postpath) == 0:
            data = {}
            data['timings'] = dict((name, self.collector.get_combined("timing:%s" % name).to_dict()) for name in self.collector.stats.timings.keys())
            data['counters'] = dict((name, self.collector.get_combined("counter:%s" % name, series=series)) for name in self.collector.stats.counters.keys())
            return respond(request, data)
        else:
            name = '/'.join(request.postpath)

            data = {}
            try:
                for n in name.split(','):
                    data[n] = self.collector.get_combined(name, series=series).to_dict()
            except KeyError:
                return respond(request, dict(code=404, error="Not Found"), code=404);

            if ',' in name:
                return respond(request, data)
            else:
                return respond(request, data[name])

GRAPH_HTML = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">
<html>
<head>
  <meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">
  <script type="text/javascript" src="http://danvk.org/dygraphs/dygraph-combined.js"></script>
  <script type="text/javascript" src="http://ajax.googleapis.com/ajax/libs/jquery/1.4.1/jquery.min.js"></script>
  <script type="text/javascript">
    if (document.location.search.length > 0) {
      $(document).ready(drawChart);
    } else {
      $.getJSON("graph_data", function(datadump) {
        var keys = datadump["keys"].sort();
        for (i in keys) {
          $("#contents").append('<a href="graph?' + keys[i] + '">' + keys[i] + '</a><br/>');
        }
        $("#graph-container").css("display", "none");
      });
    }

    function drawChart() {
      var key = document.location.search.substr(1);
      g = new Dygraph(
        document.getElementById("chart"),
        "graph_data/" + key,
        {
          //rollPeriod: 1,
          showRoller: true,
          customBars: true
          //yAxisLabelWidth: 30
        }
      );
    }
  </script>
</head>
<body>

<div id="graph-container">
<div id="chart" style="width: 640px; height: 320px;"></div>
</div>
<div id="contents"></div>

</body>
</html>
"""



import threading

from ostrich.stats_provider import StatsProvider
from ostrich.timing import Timing

class StatsCollection(StatsProvider):
    def __init__(self, counters=None, timings=None):
        self.counters = counters or {}
        self.timings = timings or {}
        self.counters_lock = threading.Lock()
        self.timings_lock = threading.Lock()
    
    def add_timing(self, name, duration):
        self.get_timing(name).add(duration)
    
    def incr(self, name, count=1):
        self.get_counter(name).incr(count)
    
    def get_counter_stats(self, reset=False):
        with self.counters_lock:
            return dict(map(lambda (k, counter): (k, counter.get(reset)), self.counters.items()))
    
    def get_timing_stats(self, reset=False):
        with self.timings_lock:
            return dict(map(lambda (k, timing): (k, timing.get(reset)), self.timings.items()))
    
    def get_counter(self, name):
        counter = self.counters.get(name, None)
        if counter is None:
            with self.counters_lock:
                counter = self.counters.get(name, None)
                if counter is None:
                    counter = Counter()
                    self.counters[name] = counter
        return counter
            
    
    def get_timing(self, name):
        timing = self.timings.get(name, None)
        if timing is None:
            with self.timings_lock:
                timing = self.timings.get(name, None)
                if timing is None:
                    timing = Timing()
                    self.timings[name] = timing
        return timing
    
    def clear_all(self):
        with self.counters_lock:
            self.counters.clear()
        with self.timings_lock:
            self.timings.clear()

class Counter(object):
    def __init__(self):
        self.value = 0;
        self.lock = threading.Lock()
    
    def incr(self, n=1):
        with self.lock:
            self.value += n
            return self.value
    
    def __call__(self):
        return self.value
    
    def reset(self):
        with self.lock:
            self.value = 0
            return self.value
    
    def get(self, reset=False):
        if reset:
            with self.lock:
                try:
                    return self.value
                finally:
                    self.value = 0
        else:
            return self.value
    
    def __eq__(self, other):
        return self.value == other.value



from bisect import bisect
import sys

class Histogram(object):
    BUCKET_OFFSETS = [1, 2, 3, 4, 5, 7, 9, 11, 14, 18, 24, 31, 40, 52, 67, 87, 113, 147, 191, 248,
          322, 418, 543, 706, 918, 1193, 1551, 2016, 2620, 3406, 4428, 5757, 7483,
          9728, 12647, 16441, 21373, 27784, 36119, 46955, 61041, 79354, 103160, 134107,
          174339, 226641, 294633, 383023, 497930, 647308, 841501, 1093951]
    
    def __init__(self, *values):
        self.num_buckets = len(self.BUCKET_OFFSETS) + 1
        self.buckets = [0] * self.num_buckets
        self.total = 0
        if values:
            for val in values:
                self.add(val)

    @classmethod
    def from_list(cls, buckets, bucket_offsets=None):
        """This method will be lossy if the bucket_offsets are different"""
        if not bucket_offsets:
            bucket_offsets = cls.BUCKET_OFFSETS
        h = cls()
        if bucket_offsets == cls.BUCKET_OFFSETS:
            for i, v in enumerate(buckets):
                h.buckets[i] = v
            h.total = sum(buckets)
        else:
            for i, v in enumerate(buckets):
                # hack because binary_search does an exclusive max
                h.add(bucket_offsets[i]-1, v)
        return h
    
    def add(self, n, count=1):
        index = bisect(self.BUCKET_OFFSETS, n)
        self.buckets[index] += count
        self.total += count
    
    def clear(self):
        for i in range(self.num_buckets):
            self.buckets[i] = 0
        self.total = 0

    def get_percentile(self, percentile):
        sum = 0
        index = 0
        while sum < percentile * self.total:
            sum += self.buckets[index]
            index += 1
        
        if index == 0:
            return 0
        elif index - 1 >= len(self.BUCKET_OFFSETS):
            return sys.maxint
        else:
            return self.BUCKET_OFFSETS[index - 1] - 1
    
    def merge(self, other):
        for i in range(self.num_buckets):
            self.buckets[i] += other.buckets[i]
        self.total += other.total
    
    def clone(self):
        new = Histogram()
        new.merge(self)
        return new
